extends Node2D

# デバッグ用print
const DBG_PRINT = false

# 無音とするdB.
# 正確には -80dBが無音だけれどフェード時間を考えて -48dB
# 気になる場合は フェード時間を短くすれば良さそう.
const MUTE_DB = -48.0
# デフォルトのフェード時間.
const FADETIME_DEFAULT = 1.0
# 音量プロパティ名
const VOLUME_DB = "volume_db"
# フェードアウト時のイージング
const FADE_OUT_EASE = Tween.TRANS_EXPO # クロスフェードっぽいカーブにする.
#const FADE_OUT_EASE = Tween.TRANS_CUBIC # 音の重なりが気になる場合はこっち.
#const FADE_OUT_EASE = Tween.TRANS_LINEAR # 音の重なりが気になる場合はこっち.

# トラックの状態.
enum eTrackState {
	NONE, # 存在しないトラック.
	STOP, # 停止中.
	MUTE, # 無音 (再生中).
	FADE_OUT, # フェードアウト.
	FADE_IN, # フェードイン.
	PLAYING, # 再生中.
}

class TrackInfo:
	var idx:int # トラック番号.
	var state:eTrackState # 状態.
	var player:AudioStreamPlayer
	var src_db:float
	var tween:Tween

var _tracks:Array[TrackInfo] = []

## 登録しているトラックの数を取得する.
func count_track() -> int:
	return _tracks.size()

## 指定のトラックが再生中かどうか.
func is_playing_track(idx:int) -> bool:
	for track in _tracks:
		if idx == track.idx:
			return track.player.playing
	return false
	
## 指定のトラックの再生位置を取得する.
func get_track_playback_pos(idx:int) -> float:
	for track in _tracks:
		if idx == track.idx:
			return track.player.get_playback_position()
	return -1	

## トラックの状態を取得する.
func get_track_state(idx:int) -> eTrackState:
	for track in _tracks:
		if idx == track.idx:
			return track.state
	return eTrackState.NONE # 存在しないトラック.
	
## トラックの状態を文字列で取得する.
func get_track_state_dbg(idx:int) -> String:
	for track in _tracks:
		if idx == track.idx:
			return eTrackState.keys()[track.state]
	return eTrackState.keys()[eTrackState.NONE] # 存在しないトラック.
	
## トラックに対応する音量を取得する.
func get_track_db(idx:int) -> float:
	for track in _tracks:
		if idx == track.idx:
			return track.player.volume_db
	return MUTE_DB

## 無音ですべて再生.
func play_all_mute() -> void:
	for info in _tracks:
		# 止めてから再生.
		_stop(info)
		info.player.play()

## すべて無音にする.
func mute_all() -> void:
	for info in _tracks:
		_stop_tween(info)
		info.player.volume_db = MUTE_DB
		

## 指定のトラックを再生する (該当しないトラックはフェードアウトする).
func play(idx:int=0, time:float=FADETIME_DEFAULT, auto_stop:bool=true) -> void:
	for info in _tracks:
		if idx == info.idx:
			# フェードイン.
			_play(info, true, time, auto_stop)
		else:
			# フェードアウト.
			_play(info, false, time, auto_stop)

## すべて止める.
func stop_all() -> void:
	for info in _tracks:
		_stop(info)

## 開始.
func _ready() -> void:
	var idx = 0 # トラック番号.
	for obj in get_children():
		if not obj is AudioStreamPlayer:
			continue
		var player:AudioStreamPlayer = obj
		# AudioStreamPlayerのみ登録する.
		var info = TrackInfo.new()
		info.idx = idx
		info.state = eTrackState.STOP # 停止中.
		info.player = player
		info.src_db = player.volume_db
		player.volume_db = MUTE_DB # 無音にしておく.
		_tracks.append(info)
		
		if DBG_PRINT:
			# デバッグ用.
			print(idx, ":", info.player.stream)
		
		idx += 1

## トラック情報の取得.
func _get_track(idx:int) -> TrackInfo:
	return _tracks[idx]

## 再生.
func _play(info:TrackInfo, is_fade_in:bool, time:float=1.0, auto_stop:bool=true) -> void:
	# Tweenが存在していたらいったん消す.
	_stop_tween(info)
	# Tweenを生成.
	info.tween = create_tween()
	var tw:Tween = info.tween
	if is_fade_in:
		# フェードイン.
		if DBG_PRINT:
			print("fade in:", info.idx)
		if info.player.playing == false:
			# フェードイン時に再生していなかったら再生.
			info.player.play()
		info.state = eTrackState.FADE_IN
		tw.tween_property(info.player, VOLUME_DB, info.src_db, time)
		# finishedシグナルで完了時のステータスを更新する.
		info.tween.finished.connect(func(): info.state = eTrackState.PLAYING)
	else:
		# フェードアウト.
		if DBG_PRINT:
			print("fade out:", info.idx)
		info.state = eTrackState.FADE_OUT
		tw.set_ease(Tween.EASE_IN)
		tw.set_trans(FADE_OUT_EASE)
		tw.tween_property(info.player, VOLUME_DB, MUTE_DB, time)
		# finishedシグナルで完了時のステータスを更新する.
		info.tween.finished.connect(
			func():
				info.state = eTrackState.MUTE
				if auto_stop:
					# 自動停止する場合は止める.
					_stop(info)
		)
	tw.play()

## 停止.
func _stop(info:TrackInfo) -> void:
	_stop_tween(info)
	info.player.stop()
	info.player.volume_db = MUTE_DB
	info.state = eTrackState.STOP

## Tweenを止める.
func _stop_tween(info:TrackInfo) -> void:
	if info.tween:
		info.tween.kill()
